ignore_regexps = [
    # ignore trivial fixes
    r'Auto-generating.*',
    r'spelling|typo',
    r'bump.*version',
    # all merged commits in the PR will appear in changelog anyway so
    # the PR merge commit is not needed.
    r'Merge pull request #\d+ from.*'
]

section_regexps = [
    ('Other', None),  ## Match all lines
]


empty_string = lambda _: ''
body_process = empty_string


subject_process = (strip |
    ReSub(r'^([cC]hg|[fF]ix|[nN]ew)\s*:\s*((dev|use?r|pkg|test|doc)\s*:\s*)?([^\n@]*)(@[a-z]+\s+)*$', r'\4') |
    ucfirst | final_dot)


tag_filter_regexp = r'^[0-9]+\.[0-9]+(\.[0-9]+)?$'


unreleased_version_label = "Upcoming release (unreleased changes)"


output_engine = rest_py
#output_engine = mustache("restructuredtext")
#output_engine = mustache("markdown")
#output_engine = makotemplate("restructuredtext")


include_merge = True
